package VM2;

import java.util.*;

/*
 * @author Derk Crezee & Daan van Schuppen UVA
 *
 * A region is a certain area of pixels in a picture
 * that have similar values. Properties of regions are
 * bounding boxes and area, centroid etc. <br>
 * A structuering element is used to check for surrounding
 * frames. On the places where the element is true, pixels
 * are added to be checked for adding to region.
 * 
 * TODO: implementing histogram information for regions
 */
public class Region {
	CameraHandler myCameraHandler;
	
	Hashtable pointsVisited = new Hashtable();
	Vector boundary = new Vector();
	
	private int numberOfPixels;
	
	// Bounnding box information
	private int xMin;
	private int xMax;
	private int yMin;
	private int yMax;
	
	// average
	private int regionAverage = 50;	
	// frame information
	private int frameSize;
	private int lineSize;
	
	// structuering element
	boolean[][] struct;
	
	byte[] color = null;
	
	private int[] averageDirection;
	private int time;
	private int averageNumber = 3;
	
	
	/*
	 * The x and y coordinates define the seed of the region
	 */
	public Region(CameraHandler myCameraHandler, int seed_x, int seed_y, boolean[][] struct) {
		
		this.averageDirection = new int[] {0,0};
		this.time = 0;
		
		numberOfPixels = 0;
		
		// Set CameraHandler
		this.myCameraHandler = myCameraHandler;
		
		// Struct
		this.struct = struct;
		
		// Set Maxima
		xMin = seed_x;
		xMax = seed_x;
		yMin = seed_y;
		yMax = seed_y;
		
		// set frame information
		frameSize = myCameraHandler.getCamera().getFrameSize();
		lineSize = myCameraHandler.getCamera().getLineSize();
		
		// Add seed to boundary array
		int[] seed = {seed_x, seed_y};
		boundary.add(seed);
	}
	/*
	 * getDirection: returns the summation of the total movements for
	 * 				 this region over 'this.time' number of frames
	 */
	public int[] getDirection() {
		return this.averageDirection;
	}
	/*
	 * setDirection: sets the direction
	 * 
	 */
	public void setDirection(int[] direction) {
		this.averageDirection = direction;
	}
	/*
	 * getTime: returns the number of frames of which the total movement
	 *          'direction' has been summed 
	 */
	public int getTime() {
		return this.time;
	}
	/* 
	 * setTime: sets the time
	 */
	public void setTime(int time) {
		this.time = time;
	}
	/*
	 * updateDirection: adds the movement of one frame to the total movement
	 * 				    if the average number of frames is exceeded, the total
	 * 					movement will be divided to get an average which will count
	 * 					for one frame
	 */
	public void updateDirection(int[] direction) {
		this.averageDirection[0]+= direction[0];
		this.averageDirection[1]+= direction[1];
		this.time++;
		if (this.time>=this.averageNumber) {
			this.averageDirection[0] =(int)( (double)this.averageDirection[0]/((double)this.time));
			this.averageDirection[1] =(int)( (double)this.averageDirection[1]/((double)this.time));
			this.time = 0;
		}
	}
	/*
	 * Recusively grows the region, adding similar pixels to the region
	 */
	public void grow() {
		
		// Do until boundary has elements
		while(!boundary.isEmpty()) {
			// Get first pixel from boundary
			int[] point = (int[])boundary.remove(0);
			
			// Add it to the hashtable if its not already there
			// Only compute if its not already
			if (!pointsVisited.containsKey(point)) {
				// Add to pointsVisited
				pointsVisited.put(point,new Integer(1));
				
				// Check for similarity and grow if similar
				if(isSimilar(point)) {
						
					numberOfPixels++;
					
					// Make point black
					removePoint(point);
					
					// Update maxima
					updateMaxima(point);
					
					// start and stop
					int xStart = point[0] - struct.length/2;
					int xStop = point[0] + struct.length/2;
					int yStart = point[1] - struct[0].length/2;
					int yStop = point[1] + struct[0].length/2;
					
					// Add surrounding pixels to boundary					
					for(int i=xStart; i <= xStop; i++) {
						for(int j=yStart; j <= yStop; j++) {
							// Has to be within frame
							if(i>=0 && i<lineSize) {
								if(j>=0 && j<frameSize) {
									int[] tmp = {i,j};
									// Only points indicated by struct
									if(struct[i-xStart][j-yStart]) {								
										boundary.add(tmp);
									} else {
										removePoint(tmp);
									}
								}
							} 
						}						
					}
				}
			}
			else {
				// remove point from boundary
				boundary.removeElement(point);
			}
		}
	}
	
	/*
	 * Checks wether point is similar to region points
	 */
	private boolean isSimilar(int[] point) {
		// Set buffer
		myCameraHandler.setBuffer(point[0], point[1]);
		
		// Get pixel
		byte[] pixel = new byte[3];
		myCameraHandler.getBufferRGB(pixel);
		
		// Take average
		int average = ( (pixel[0] & 0xFF) + (pixel[1] & 0xFF) + (pixel[2] & 0xFF) )/3;
		
		// Check for similarity
		if (average >= regionAverage) {
			return true;
		}
		return false;
	}
	
	/*
	 * Updates maxima where needed
	 */
	private void updateMaxima(int[] point) {
		if(point[0] < xMin) {
			xMin = point[0];
		}
		if(point[0] > xMax) {
			xMax = point[0];
		}
		if(point[1] < yMin) {
			yMin = point[1];
		}
		if(point[1] > yMax) {
			yMax = point[1];
		}
	}
	
	/*
	 * Removes region from picture
	 */
	private void removePoint(int[] point) {
		// set buffer
		myCameraHandler.setBuffer(point[0], point[1]);
		
		// Make picture black
		byte[] pixel = new byte[] {0,0,0};
		//myCameraHandler.getBufferRGB(pixel);
		
		myCameraHandler.setBufferRGB(pixel);
	}
	
	public void setColor(byte[] color) {
		this.color = color;
		
	}
	public byte[] getColor() {
		return this.color;
	}
	/*
	 * Draws bounding box in picture
	 */
	public void drawBBox() {
		if (this.color!=null){
			for(int i=xMin; i<=xMax; i++ ) {
				myCameraHandler.setBuffer(i,yMin);		
				myCameraHandler.setBufferRGB(this.color);
				
				myCameraHandler.setBuffer(i,yMax);		
				myCameraHandler.setBufferRGB(this.color);
			}
			
			for(int i=yMin; i<=yMax; i++ ) {
				myCameraHandler.setBuffer(xMin,i);		
				myCameraHandler.setBufferRGB(this.color);
				
				myCameraHandler.setBuffer(xMax,i);		
				myCameraHandler.setBufferRGB(this.color);
			}
		}
	}
	
	/*
	 * Returns the area of the BBox
	 */
	public int getArea() {
		return (xMax-xMin)*(yMax-yMin);
	}
	public int getNumberOfPixels() {
		return numberOfPixels;
	}
	public int[][] getCornerPositions() {
		int[][] tmp = new int[4][2];
		tmp[0][0] = xMin;
		tmp[0][1] = yMin;
		tmp[1][0] = xMax;
		tmp[1][1] = yMin;
		tmp[2][0] = xMin;
		tmp[2][1] = yMax;
		tmp[3][0] = xMax;
		tmp[3][1] = yMax;
		return tmp;
		
	}
	
	/*
	 * Returns the centroid of the BBox
	 */
	public int[] getCentroid() {
		int[] centroid = {xMin+(xMax-xMin)/2, yMin+(yMax-yMin)/2};
		return centroid;
	}
	/*
	 * Draws an arrow from the center of the bounding box to direction of 'direction'
	 */
	public void drawLine() {
		int x = ((int)(xMin+(xMax-xMin)/2));
		int y = ((int)(yMin+(yMax-yMin)/2));
		
		int[] direction = new int[2];
		
		direction[0] = x + (int)((double)averageDirection[0]/(double)time);
		direction[1] = y + (int)((double)averageDirection[1]/(double)time);
		
		myCameraHandler.drawLine(x,y,direction[0],direction[1]);
		
	}
}