package VM2;

import javax.media.*;
import javax.media.format.*;
import javax.media.protocol.*;
import java.io.IOException;
import java.util.Vector;
import javax.media.protocol.DataSource;

import GUI2.Stdout;
import VM2.DeviceInfo;

public class USBCam implements BufferTransferHandler, Camera {
	private DataSource videoDataSource = null; // This is the USB data source
											   // obtained from the medialocator
	private Processor processor = null; // The processor will convert RAW format
										// to RGB
	private DataSource processorOutput; // This source registers the
										// BufferTransferHandler
	private Buffer readBuffer; // This object will be used and recycled when
							   // reading data from the processorstream
	private byte[] buffer; // The byte array obtained from readBuffer

	private byte frameStatus = FRAME_IDLE;
	//	private int myFrameSize = 144;
	//	private int myLineSize = 176;
	private int myFrameSize = 480;
	private int myLineSize = 640;
//		private int myFrameSize = 288;
//		private int myLineSize = 352;

	public USBCam() {
		readBuffer = new Buffer();
		buffer = new byte[myFrameSize * myLineSize * 3];
		readBuffer.setData(new byte[buffer.length]); // Ensure readBuffer has
													 // data right away

		///////////////////////////////////////////
		// Find all available devices and select //
		// the default videop capture device //
		///////////////////////////////////////////

		boolean debugDeviceList = true;

		Stdout.log("get list of all media devices ...");
		Vector deviceListVector = CaptureDeviceManager.getDeviceList(null);
		if (deviceListVector == null) {
			Stdout
					.log("... error: media device list vector is null, program aborted");
			System.exit(0);
		}
		if (deviceListVector.size() == 0) {
			Stdout
					.log("... error: media device list vector size is 0, program aborted");
			System.exit(0);
		}

		CaptureDeviceInfo deviceInfo = null; // To iterate over all available
											 // devices
		CaptureDeviceInfo captureVideoDevice = null; // To memorize the default
													 // video capture device
		VideoFormat captureVideoFormat = null;
		Format deviceFormat[] = null;
		MediaLocator videoMediaLocator = null;

		for (int x = 0; x < deviceListVector.size(); x++) {
			// display device name
			deviceInfo = (CaptureDeviceInfo) deviceListVector.get(x);
			if (debugDeviceList)
				Stdout.log("device " + x + ":  " + deviceInfo.getName());

			// display device formats
			deviceFormat = deviceInfo.getFormats();

			for (int y = 0; y < deviceFormat.length; y++) {
				// search for default video device
				if (captureVideoDevice == null) {
					if (deviceFormat[y] instanceof VideoFormat) {
						captureVideoDevice = deviceInfo;
						videoMediaLocator = captureVideoDevice.getLocator();
						Stdout.log(">> Device = "
								+ captureVideoDevice.getName());
						Stdout.log(">> Locator = "
								+ videoMediaLocator.toString());
					}
				}

				// search for default video format
				if (captureVideoDevice == deviceInfo) {
					String dfs = DeviceInfo.formatToString(deviceFormat[y]);
					String indexText = myLineSize + "x" + myFrameSize;
					if ((dfs.indexOf(indexText) > 0)
							&& (dfs.indexOf("rgb") > 0)) {
						captureVideoFormat = (VideoFormat) deviceFormat[y];
						Stdout.log(">> Format = "
								+ DeviceInfo.formatToString(deviceFormat[y]));
					}
				}
				if (debugDeviceList)
					Stdout.log(" - format:  "
							+ DeviceInfo.formatToString(deviceFormat[y]));
			}
		}
		if (captureVideoDevice == null) {
			Stdout.log("No video device found");
			System.exit(0);
		}
		Stdout.log("... list completed.");
		Stdout.log("Setting up data source...");

		try {
			videoDataSource = Manager.createDataSource(videoMediaLocator);
		} catch (NoDataSourceException nse) {
			Stdout.logAndAbortException(nse);
		} catch (Exception e) {
			Stdout.logAndAbortException(e);
		}

		if (!DeviceInfo.setFormat(videoDataSource, captureVideoFormat)) {
			Stdout.log("Error: unable to set video format - program aborted");
			System.exit(0);
		}
		Stdout.log("Setting up data formats...");

		///////////////////////////////////////////////
		// Setup the processor that will convert raw //
		// video data to RGB //
		///////////////////////////////////////////////

		Stdout.log("Creating Processor...");

		try {
			processor = Manager.createRealizedProcessor(new ProcessorModel(
					videoDataSource, new Format[]{new VideoFormat(
							VideoFormat.RGB)}, new FileTypeDescriptor(
							FileTypeDescriptor.RAW)));
			processorOutput = processor.getDataOutput();
			processorOutput.start();
			PushBufferStream[] pushStrms = ((PushBufferDataSource) processorOutput)
					.getStreams();

			// Set the transfer handler to receive pushed data from
			// the push DataSource.
			for (int i = 0; i < pushStrms.length; i++) {
				pushStrms[i].setTransferHandler(this);
			}
		} catch (Exception e) {
			Stdout.logAndAbortException(e);
		}

		processor.configure();
		waitForState(Processor.Realized);
		Stdout.log("Processor realized...");
		Stdout.log("Camera Ready...");
	}

	private void waitForState(int state) {
		try {
			while (processor.getState() != state)
				wait();
		} catch (Exception e) {
		}
	}

	public void start() {
		try {
			Stdout.log("Prepare capture..");
			videoDataSource.start();
		} catch (Exception e) {
			Stdout.logAndAbortException(e);
		}

		Stdout.log("prefetching..");
		processor.prefetch();
		waitForState(Processor.Prefetched);

		Stdout.log("starting..");
		processor.start();
		waitForState(Processor.Started);

		Stdout.log("Started");
	}

	public void stop() {
		if (videoDataSource != null) {
			Stdout.log("Stop capture...");
			try {
				videoDataSource.stop();
			} catch (IOException e) {
				System.err.println("Error stopping source!\n" + e.getMessage());
			}
			Stdout.log("Done");
		}
	}

	public void dispose() {
		stop();
		processor.stop();
		try {
			processorOutput.stop();
		} catch (IOException e) {
			e.printStackTrace();
		}

		Stdout.log("Disconnecting...");
		videoDataSource.disconnect();
	}

	/**
	 * This will get called when there's data pushed from the
	 * PushBufferDataSource. Here is were we get our frames from the USB camera
	 * and put them in the byte array 'buffer'.
	 */
	public synchronized void transferData(PushBufferStream stream) {
		// If the readbuffer is being used at this time, wait until it's free
		// Then signal that we are altering the buffer
		try {
			while (!frameStatus(FRAME_IDLE, FRAME_UPDATING))
				wait();
		} catch (Exception e) {
		}

		// Get the frame from the stream
		try {
			stream.read(readBuffer);
		} catch (IOException e) {
			System.err.println(e);
		}

		// The buffer has been updated. Signal it's okay to use the readBuffer
		// again.
		frameStatus(FRAME_UPDATING, FRAME_IDLE);
	}

	private static final byte FRAME_IDLE = 0;
	private static final byte FRAME_UPDATING = 1;
	private static final byte FRAME_COPYING = 2;

	private synchronized boolean frameStatus(byte condition, byte value) {
		if (frameStatus == condition) {
			frameStatus = value;
			return true;
		}
		return false;
	}

	/**
	 * Gets a new frame and pushes the old frame into the previous buffer It
	 * then get the processed output buffer and returns this to display
	 */
	public synchronized byte[] getNextFrame() {
		// Check to see if the readBuffer is in use. Wait until it's free
		// and signal that we want to copy the buffer now
		try {
			while (!frameStatus(FRAME_IDLE, FRAME_COPYING))
				wait();
		} catch (Exception e) {
		}

		// Get the data from the buffer. The data is in BGR format and upside down.
		// We change it to RGB while copying the content and flipping the image over.
		if (readBuffer.isEOM())
			System.err.println("  Got EOM!");
		byte[] readData = (byte[]) readBuffer.getData();

		int yInv;
		for (int y = 0; y < myFrameSize; y++) {
			yInv = myFrameSize - y - 1;
			for (int x = 0; x < myLineSize; x++) {
				buffer[(y * myLineSize + x) * 3 + 0] = readData[(yInv
						* myLineSize + x) * 3 + 2];
				buffer[(y * myLineSize + x) * 3 + 1] = readData[(yInv
						* myLineSize + x) * 3 + 1];
				buffer[(y * myLineSize + x) * 3 + 2] = readData[(yInv
						* myLineSize + x) * 3 + 0];
			}
		}

		// We're done copying the buffer, so release the readBuffer now.
		frameStatus(FRAME_COPYING, FRAME_IDLE);
		return buffer;
	}

	public byte[] getCurrentFrame() {
		// only return this buffer if it is not being updated
		try {
			while (frameStatus == FRAME_COPYING)
				wait();
		} catch (Exception e) {
		}

		return buffer;
	}

	public byte getType() {
		return Camera.TYPE_USB;
	}

	public int getFrameSize() {
		return myFrameSize;
	}

	public int getLineSize() {
		return myLineSize;
	}
}