package VM2;

/*
 * @author Derk Crezee & Daan van Schuppen UVA
 *
 * regionMotions: this class is able to detect motions in two succesive frames 
 * 				  (after background subtraction, opening, closing) 
 * 				  motion is detected by matching the regions of the current frame
 * 				  and the current frame. Matching of regions is done by a Nearest Neighbor
 * 				  algorithm based on the Eucledian distance between two regions and the
 * 				  size difference between two regions
 */
import java.util.*;

public class regionMotions {
	
	CameraHandler myCameraHandler;
	// definition of the structuring element
	public static final int
			CORNER_3X3 = 0,
			SQUARE_3X3 = 1,
			CORNER_5X5 = 2,
			CORNER_9X9 = 3;
	
	// the regions (bounding boxes) of the previous frame
	Vector previousSeeds = new Vector();
	// the regions (bounding boxes) of the current frame
	Vector currentSeeds = new Vector();
	
	// minimal size to be a bounding box
	private int minimalSize;
	// maximum distance from a bounding box in the previous frame
	// to a bounding box in the current frame to become similar
	private int maxDistance;
	
	// structuring element used in the region class
	boolean[][] struct;
	
	/*
	 * - myCameraHandler: CameraHandler
	 * - minimalSize: minimal size to be a bounding box
	 * - maxDistance: maximum distance from a bounding box in the previous frame
	 *   			  to a bounding box in the current frame to become similar
	 * - structuringElement: structuring element used by the region class
	 */
	public regionMotions(CameraHandler myCameraHandler,int minimalSize,int maxDistance,int structuringElement) {
		this.myCameraHandler = myCameraHandler;
		this.minimalSize = minimalSize;
		this.maxDistance = maxDistance;
		previousSeeds = new Vector();
		currentSeeds = new Vector();
		setStructuringElement(structuringElement);
	}
	
	public regionMotions(CameraHandler myCameraHandler) {
		this.myCameraHandler = myCameraHandler;
		previousSeeds = new Vector();
		currentSeeds = new Vector();
		setStructuringElement(0);
	}
	/* 
	 * findBoundingBoxes: finds the different regions in the current frame
	 * 					  by looping through the image, each time a white pixel is
	 * 					  found a region is set up for this pixel
	 */
	public void findBoundingBoxes() {
		
		int i,iMin,iMax,j,jMin,jMax;
		int blockSize = 1;
		int stepSize = 1;
				
		iMin = Math.round(blockSize/2);
		iMax = myCameraHandler.getCamera().getFrameSize() - Math.round(blockSize/2);
	
		jMin = iMin;
		jMax = myCameraHandler.getCamera().getLineSize() - Math.round(blockSize/2);
	
		myCameraHandler.setSourceBuffer(CameraHandler.BUFFERID_OUT);
		myCameraHandler.setDestinationBuffer(CameraHandler.BUFFERID_OUT);
		
		currentSeeds = new Vector();
		
		//loop through the current frame and find the regions
		for (i=iMin; i<iMax; i=i+stepSize)
		{
			for (j=jMin; j<jMax; j=j+stepSize)
			{	
				// RGB values of current pixel
				int[] RGB = myCameraHandler.getRGB(j,i);
				// if white
				if ((RGB[0]==255)&&(RGB[0]==255)&&(RGB[0]==255)) {
					// start a new region and grow it according to structuring element struct
					Region tmp = new Region(myCameraHandler,j,i, struct);
					tmp.grow();
					
					if (tmp.getArea()>this.minimalSize) {
						currentSeeds.add(tmp);
					}
				}
			}
		}
		
		
		
		// determine the motions between the regions of the 
		determineMotion();

		// set the background to the source camera in image
		myCameraHandler.setSourceBuffer(CameraHandler.BUFFERID_IN);
		myCameraHandler.setDestinationBuffer(CameraHandler.BUFFERID_OUT);
		myCameraHandler.bufferCopy();
		// show the current bounding boxes
		show();
	}
	/*
	 * determineMotion: determines the motions between the different regions
	 *                  of the current frame and the regions of the previous
	 * 					frame
	 */
	public void determineMotion() {
		Region currentSeed, previousSeed;
		int[] currentPos, previousPos;
		double distance;
		double difSize;
		
		// loop through the current regions and calculate for each the distances
		// and the size differences of the previous frame regions
		for (int i=0;i<currentSeeds.size();i++) {
			
			currentSeed = (Region) currentSeeds.elementAt(i);
			currentPos = currentSeed.getCentroid();
			
			double[] distances = new double[previousSeeds.size()];
			double[] sizes = new double[previousSeeds.size()];
			int[] rankDistances = new int[previousSeeds.size()];
			int[] rankSizes = new int[previousSeeds.size()];
			
			// loop through the previous regions and calculate the distance and size 
			// difference between the current region and the previous regions
			for (int j=0;j<previousSeeds.size();j++) {
				previousSeed = (Region) previousSeeds.elementAt(j);
				previousPos = previousSeed.getCentroid();
				// Eucledian distance
				distance =Math.sqrt((previousPos[0]-currentPos[0])*(previousPos[0]-currentPos[0]) + (previousPos[1]-currentPos[1])*(previousPos[1]-currentPos[1]));
				// difference in size
				difSize = Math.sqrt( (previousSeed.getNumberOfPixels() - currentSeed.getNumberOfPixels()) * (previousSeed.getNumberOfPixels() - currentSeed.getNumberOfPixels()) );
				
				distances[j] = distance;
				sizes[j] = difSize;
			}
			// get the ranking of the current region with the previous regions with respect
			// to the distance
			rankDistances = getRanking(distances);
			// get the ranking of the current region with the previous regions with respect
			// to the size
			rankSizes = getRanking(sizes);
			int min = 999999;
			int minIndex=-1;
			
			// find the previous region which is closest to the current region with respect
			// to distance and size
			for (int j=0;j<rankDistances.length;j++) {
				
				if (rankSizes[j]+rankDistances[j]<min) {
					min = rankSizes[j]+rankDistances[j];
					minIndex = j;
				} else if (rankSizes[j]+rankDistances[j]==min) {
					// this is the case if the rankings of two previous regions are the same
					// in this case take the one with the lowest rank in size and distance
					if (distances[j] + sizes[j] < distances[minIndex] + sizes[minIndex]) {
						min = rankSizes[j]+rankDistances[j];
						minIndex = j;
					}
				}
			}
			// a previous region match is found
			if (minIndex!=-1) {
				int[] direction = new int[2];
				Region tempCurrent = (Region)currentSeeds.elementAt(i);
				Region tempPrevious = (Region)previousSeeds.elementAt(minIndex);
				int[] centroidCurrent = tempCurrent.getCentroid();
				int[] centroidPrevious = tempPrevious.getCentroid();
				direction[0] = centroidCurrent[0] - centroidPrevious[0];
				direction[1] = centroidCurrent[1] - centroidPrevious[1];
				// distance between previous and current region
				int dist =(int) Math.sqrt(direction[0]*direction[0] + direction[1]*direction[1]);
				
				// the distance between the matched previous region and the current region is small enough
				// to be a real match
				if ( dist < maxDistance ) {
					// take over the color of the matched previous region 
					// and take over the average movement so far
					if (tempPrevious.getColor()!=null) {
						tempCurrent.setColor(tempPrevious.getColor());
						tempCurrent.setTime(tempPrevious.getTime());
						tempCurrent.setDirection(tempPrevious.getDirection());
						// draw a line from the center of the region to the direction of movement
						// this direction is an average over a particular number of frames
						tempCurrent.updateDirection(direction);
					} else {
						// set to a random color
						byte[] bColor = {(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255)};
						tempCurrent.setColor(bColor);
						
					}
				// the distance between the matched previous region and the current region are too far to match
				} else {
					// set the current region to a random color
					byte[] bColor = {(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255)};
					((Region)currentSeeds.elementAt(i)).setColor(bColor);
					
				}
				
			} else {
				// no match is found set the region to a random color
				byte[] bColor = {(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255),(byte)(int)(Math.random()*255)};
				((Region)currentSeeds.elementAt(i)).setColor(bColor);
			}
		}
		// set the previous regions to the current regions
		previousSeeds = (Vector)currentSeeds.clone();
	}
	/*
	 * getRanking: sorts a double array toBeSorted and ranks the indexes of the original
	 *             array according to the sorted array
	 */
	public int[] getRanking(double[] toBeSorted) {
		double[] sort = (double[])toBeSorted.clone();
		boolean sorted = false;
		
		int[] ranking = new int[sort.length];
		int tmp1=0;
		double current,previous=0;
		double tmp2=0;
		// set the original indexes
		for (int i=0;i < ranking.length;i++) {
			ranking[i] = i;
		}
		while (!sorted) {
			sorted = true;
			for (int i=0; i<sort.length;i++) {
				if (i==0) {
					previous = sort[i];
				} else {
					current = sort[i];
					if (current < previous) {
						// switch the double array sort and switch the indexes
						sorted=false;
						tmp1 = ranking[i-1];
						ranking[i-1] = ranking[i];
						ranking[i] = tmp1;
						tmp2 = sort[i-1];
						sort[i-1] = sort[i];
						sort[i] = tmp2;
					} else {
						previous = current;
					}
				}
			}
		}
		int[] test = (int[])ranking.clone();
		for(int i=0; i<test.length; i++) {
			int index = test[i];
			ranking[index] = i;
		}
		return ranking;
	}
	/* 
	 * show: shows the bounding boxes of the current regions
	 */
	public void show() {
		for (int q=0;q<currentSeeds.size();q++) {
			Region tmp = (Region)currentSeeds.elementAt(q);
			byte[] red = {(byte)255,0,0};
			tmp.drawBBox();
			tmp.drawLine();
		}
		
	}
	/*
	 * setStructuringElement: sets the structuring element which is used
	 * 						  to grow a region
	 * 						  true in the matrix means that the pixel of 
	 * 						  that location is used to determine the region
	 * 						  false means that the pixel is not used for
	 * 						  determination of a region
	 */
	public void setStructuringElement(int ID) {
		switch (ID) {
			case CORNER_3X3 :
				struct = new boolean[][] {{true,false,true},
										  {false,false,false},
										  {true,false,true}};
				break;
			case SQUARE_3X3 :
				struct = new boolean[][]{ {true,true,true},
					       				  {true,false,true},
										  {true,true,true}};
				break;

			case CORNER_5X5 :
				struct = new boolean[][]{ {true,false,false,false,true},
										  {false,false,false,false,false},
										  {false,false,false,false,false},
										  {false,false,false,false,false},
										  {true,false,false,false,true}};
				break;
			case CORNER_9X9 :
				struct = new boolean[][]{ {true,false,false,false,false,false,false,false,true},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {false,false,false,false,false,false,false,false,false},
										  {true,false,false,false,false,false,false,false,true}};
				break;
			default :
				System.out.println("Unknown structuring element in regionMotions: "
								+ ID);
				struct = new boolean[][] {{true,false,true},
										  {false,false,false},
										  {true,false,true}};
				break;
		}
	}
}